/*
Part of the Bungie.net Myth2 Metaserver source code
Copyright (c) 1997-2002 Bungie Studios
Refer to the file "License.txt" for details
*/

#include "cseries.h"
#include "byte_swapping.h"

#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>
#include <time.h>

#include "metaserver_common_structs.h"
#include "stats.h"
#include "authentication.h"
#include "bungie_net_player.h"
#include "bungie_net_order.h"
#include "users.h"
#include "rank.h"
#include "games.h"
#include "metaserver_packets.h"
#include "room_packets.h"
#include "server_code.h"
#include "remote_commands.h"
#include "room_globals.h"
#include "environment.h"

#include "game_search_packets.h"


/* ------- structures */

/* ------- local prototypes */
static boolean try_and_handle_remote(struct client_data *client, char *message, unsigned long selected_player_id);

/* ------- code */
boolean handle_remote_command(
	struct client_data *client, 
	char *message, 
	short length)
{
	boolean handled= FALSE;
	
	if(length>sizeof(struct myth_metaserver_chat_packet))
	{
		struct myth_metaserver_chat_packet *packet= (struct myth_metaserver_chat_packet *) message;
		short type= packet->type;
		unsigned long selected_player_id= packet->selected_player_id;
		
#ifdef little_endian
		type= SWAP2(type);
		selected_player_id= SWAP4(packet->selected_player_id);
#endif

		switch(type)
		{
			case _myth_metaserver_chat_packet:
				{
					char *user= message+sizeof(struct myth_metaserver_chat_packet);
					char *chat= user+strlen(user)+1;
					
					if(chat[0]=='.')
					{
						handled= try_and_handle_remote(client, chat, selected_player_id);
					}
				}
				break;
			
			default:
				break;
		}
	} 
	
	return handled;
}

/* ------- local code */

static void handle_help(struct client_data *client, char *rest, unsigned long selected_player_id);
static void handle_time(struct client_data *client, char *rest, unsigned long selected_player_id);
#ifndef BN2_DEMOVERSION
static void handle_caste(struct client_data *client, char *rest, unsigned long selected_player_id);
#endif
static void handle_message(struct client_data *client, char *rest, unsigned long selected_player_id);
static void handle_gag(struct client_data *client, char *rest, unsigned long selected_player_id);
static void handle_kick(struct client_data *client, char *rest, unsigned long selected_player_id);
static void handle_whereami(struct client_data *client, char *rest, unsigned long selected_player_id);
static void handle_gameinfo(struct client_data *client, char *rest, unsigned long selected_player_id);

/* external functions*/
void change_room_motd(char *motd);
struct game_data *get_first_game(void);

// These should correspond to the names found in the Myth client's "internal metaserver strings"
// string list, appearing right after the string "Prefer No Enemy Visibility".
const char *room_names[] =
{
  "Ash",
  "Bagrada",
  "Brayle",
  "Forest Heart",
  "Gonen's Bridge",
  "Llancarfan",
  "Meander",
  "Muirthemne",
  "Scales",
  "Shiver",
  "Sons of Myrgard",
  "Haven",
  "Sanctuary",
  "The Twelve Duns",
  "Soulblighter",
  "The Barrier",
  "The Blind Steppes",
  "The Deceiver",
  "Drowned Kingdom",
  "The Forge",
  "The Great Library",
  "The Summoner",
  "The Tain",
  "The Watcher",
  "Vigilance",
  "Willow Creek",
  "Tandem",
  "Legends",
  "The Comet",
  "The Cloudspine"
};

struct {
	char *command;
	char *description;
	boolean admin_only;
	void (*handler)(struct client_data *client, char *rest, unsigned long selected_player_id);
	// new commands must be added to this array - this will tell the server
	// which function to call for what command and it will also make
	// the command show up when '.help' is used
} remote_commands[]= {
	{ ".help", "Prints this text", FALSE, handle_help },
	{ ".time", "Tells the server's current time", FALSE, handle_time },
#ifndef BN2_DEMOVERSION
	{ ".caste", "Gives the caste of the selected player", FALSE, handle_caste },
#endif
	{ ".message", "Sends selected player a private message", FALSE, handle_message },
	{ ".gag", "Gag/ungag selected player - use '.gag off' to un-gag a player", TRUE, handle_gag },
	{ ".kick", "Kick selected user from the server", TRUE, handle_kick },
	{ ".whereami", "Shows the name of the room you are in", FALSE, handle_whereami },
	{ ".gameinfo", "Shows information on the first game in the room (experimental)", FALSE, handle_gameinfo }
};
#define NUMBER_OF_REMOTE_COMMANDS (sizeof(remote_commands)/sizeof(remote_commands[0]))

static boolean try_and_handle_remote(
	struct client_data *client, 
	char *message, 
	unsigned long selected_player_id)
{
	short index;
	boolean handled= FALSE;
	
	for(index= 0; index<NUMBER_OF_REMOTE_COMMANDS; ++index)
	{
		short length= strlen(remote_commands[index].command);

		if (memcmp(remote_commands[index].command, message, length) == 0)
		{
			if (remote_commands[index].admin_only == TRUE)
			{
				if (IS_PLAYER_ADMIN(client) == FALSE)
				{
					return FALSE;
				}
			}

			remote_commands[index].handler(client, message+length, selected_player_id);
			handled= TRUE;
			break;
		}
	}
	
	return handled;
}

/* ----------- handlers */
static void handle_help(
	struct client_data *client, 
	char *rest, 
	unsigned long selected_player_id)
{
        int i;
	char *full_command;
	char *separator_string = " - ";
	if(rest && rest[0] != '\0')
	{
		send_room_message_to_player(client, "I'm sorry, I didn't understand your command. Please try again.");
		return;
	}
	for(i = 0; i < NUMBER_OF_REMOTE_COMMANDS; i++)
	{
	    // skip admin only commands if user is not an admin
	    if(remote_commands[i].admin_only == TRUE)
	    {
	        if (IS_PLAYER_ADMIN(client) == FALSE)
		{
		    continue;
		}
	    }
	    full_command = (char *)malloc((strlen(remote_commands[i].command) + strlen(remote_commands[i].description) + strlen(separator_string) + 1) * sizeof(char));
	    strcpy(full_command, remote_commands[i].command);
	    strcat(full_command, separator_string);
	    strcat(full_command, remote_commands[i].description);
	    send_room_message_to_player(client, full_command);
	    free(full_command);
	}
	return;
}
static void handle_time(
	struct client_data *client, 
	char *rest, 
	unsigned long selected_player_id)
{
        time_t current_time;
	struct tm *now;
	char buffer[256];

	current_time = time(NULL);
	now = localtime(&current_time);
	strftime(buffer, sizeof(buffer), "%H:%M:%S %Z", now);

	if(rest && rest[0] != '\0')
	{
		send_room_message_to_player(client, "I'm sorry, I didn't understand your command. Please try again.");
		return;
	}
	send_room_message_to_player(client, buffer);
	
	return;
}

#ifndef BN2_DEMOVERSION
static void handle_caste(
	struct client_data *client, 
	char *rest, 
	unsigned long selected_player_id)
{
	struct client_data *selected_client;
	char buffer[512];
	char *caste_desc;
        
	if((selected_client = find_client_from_player_id(selected_player_id)) == NULL)
	{
	    send_room_message_to_player(client, "You must first select a player by clicking on their name.");
	    return;
	}

	/* set up caste description */
	switch(selected_client->caste)
	{
	    case _dagger:
	        caste_desc = "Dagger";
		break;
	    case _dagger_with_hilt:
	        caste_desc = "Dagger with Hilt";
		break;
	    case _kris_knife:
	        caste_desc = "Kris Knife";
		break;
	    case _sword_and_dagger:
	        caste_desc = "Sword and Dagger";
		break;
	    case _crossed_swords:
	        caste_desc = "Crossed Swords";
		break;
	    case _shield:
	        caste_desc = "Shield";
		break;
	    case _shield_crossed_swords:
	        caste_desc = "Shield Crossed Swords";
		break;
	    case _shield_crossed_axes:
	        caste_desc = "Shield Crossed Axes";
		break;
	    case _simple_crown:
	        caste_desc = "Simple Crown";
		break;
	    case _crown:
	        caste_desc = "Crown";
		break;
	    case _nice_crown:
	        caste_desc = "Imperial Crown";
		break;
	    case _eclipsed_moon:
	        caste_desc = "Eclipsed Moon";
		break;
	    case _moon:
	        caste_desc = "Moon";
		break;
	    case _eclipsed_sun:
	        caste_desc = "Eclipsed Sun";
		break;
	    case _sun:
	        caste_desc = "Sun";
		break;
	    case _comet:
	        caste_desc = "Comet";
		break;
	    default:
	        caste_desc = "Unknown Caste";
	}

	sprintf(buffer, "User's caste: %d - %s", selected_client->caste, caste_desc);

	if(rest && rest[0] != '\0')
	{
		send_room_message_to_player(client, "I'm sorry, I didn't understand your command. Please try again.");
		return;
	}
	send_room_message_to_player(client, buffer);
	
	return;
}
#endif

static void handle_message(
	struct client_data *client, 
	char *rest, 
	unsigned long selected_player_id)
{
	// check that there is a valid player selected
	if(find_client_from_player_id(selected_player_id) == NULL)
	{
	    send_room_message_to_player(client, "You must first select a player by clicking on their name.");
	    return;
	}

	send_chat_message(client, selected_player_id, rest);
	return;
}
static void handle_kick(
	struct client_data *client, 
	char *rest, 
	unsigned long selected_player_id)
{
	struct client_data *kick_client;

	if((kick_client = find_client_from_player_id(selected_player_id)) == NULL)
	{
	    send_room_message_to_player(client, "You must first select a player by clicking on their name.");
	    return;
	}

	if(boot_player(kick_client, rest))
	{
	    send_room_message_to_player(client, "Player was successfully kicked!");
	}
	else
	{
	    send_room_message_to_player(client, "Error kicking player - put on your steel-toed boots first.");
	}
	
	return;
}
static void handle_gag(
	struct client_data *client, 
	char *rest, 
	unsigned long selected_player_id)
{
	struct client_data *gag_client;
	boolean do_gag = TRUE;

	if((gag_client = find_client_from_player_id(selected_player_id)) == NULL)
	{
	    send_room_message_to_player(client, "You must first select a player by clicking on their name.");
	    return;
	}

	if((strcmp(rest, " off")) == 0)
	{
	    do_gag = FALSE;
	}
	if(gag_player(gag_client, do_gag))
	{
	    if(do_gag)
	    {
	        send_room_message_to_player(client, "Player was successfully gagged!");
	    }
	    else
	    {
	        send_room_message_to_player(client, "Player was successfully un-gagged!");
	    }
	}
	else
	{
	    send_room_message_to_player(client, "Error gagging player - their mouth is too big.");
	}
	
	return;
}

static void handle_whereami(
	struct client_data *client, 
	char *rest, 
	unsigned long selected_player_id)
{
	char buffer[512];

	snprintf(buffer, sizeof(buffer), "You are currently in room %s.", room_names[room_globals.room_identifier]);
	send_room_message_to_player(client, buffer);
}

static void handle_gameinfo(
	struct client_data *client, 
	char *rest, 
	unsigned long selected_player_id)
{
	char buffer[512];
	struct game_data *game = get_first_game();

	if (!game)
	{
		send_room_message_to_player(client, "There are no available games in this room.");
		return;
	}

	snprintf(buffer, sizeof(buffer), "Game name: %s", GET_GAME_NAME(game));
	send_room_message_to_player(client, buffer);
	snprintf(buffer, sizeof(buffer), "Game map: %s", GET_GAME_MAP(game));
	send_room_message_to_player(client, buffer);
	snprintf(buffer, sizeof(buffer), "Game created by: %s", GET_CLIENT_SCREEN_NAME(find_client_from_player_id(game->creator_player_id)));
	send_room_message_to_player(client, buffer);
}

