/*
Part of the Bungie.net Myth2 Metaserver source code
Copyright (c) 1997-2002 Bungie Studios
Refer to the file "License.txt" for details
*/

 /*
  * The metaserver code changes that fall outside the original Bungie.net metaserver code 
  * license were written and are copyright 2002, 2003 of the following individuals:
  *
  * Copyright (c) 2002, 2003 Alan Wagner
  * Copyright (c) 2002 Vishvananda Ishaya
  * Copyright (c) 2003 Bill Keirstead
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions
  * are met:
  *
  * 1. Redistributions of source code must retain the above copyright
  *    notice, this list of conditions and the following disclaimer.
  *
  * 2. Redistributions in binary form must reproduce the above copyright
  *    notice, this list of conditions and the following disclaimer in the
  *    documentation and/or other materials provided with the distribution.
  *
  * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
  * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
  * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
  * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
  * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
  * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
  * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
  * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
  * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
  * SUCH DAMAGE.
  *
  */

#include "cseries.h"
#include <string.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <sys/ioctl.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <stdio.h>
#include <unistd.h>
#include <netdb.h>
#include <time.h>
#include <errno.h>
#include <signal.h>

#include "environment.h"
#include "metaserver_common_structs.h"
#include "stats.h"
#include "authentication.h"
#include "bungie_net_player.h"
#include "bungie_net_order.h"
#include "users.h"
#include "room_list_file.h"
#include "games.h"
#include "rank.h"
#include "metaserver_packets.h"
#include "metaserver_codes.h"
#include "room_packets.h"
#include "server_code.h"
#include "room_globals.h"
#include "network_queues.h"
#include "game_search_packets.h"
#include "games_log.h"

enum {
	ROOM_PLAYER_LIMIT_SLOP= 15,
	DEFAULT_MAXIMUM_PLAYERS_PER_ROOM= 40,
	DEFAULT_SECONDS_BEFORE_UPDATE= 30,

	MAXIMUM_OUTSTANDING_REQUESTS= 32,

#ifdef MULTI_ROOM
	SECONDS_TO_WAIT_ON_SELECT= 0,
	MICROSECONDS_TO_WAIT_ON_SELECT= 100,
#else
	SECONDS_TO_WAIT_ON_SELECT= 10,
	MICROSECONDS_TO_WAIT_ON_SELECT= 0,
#endif

	STDIN_SOCKET= 0,
	MAXIMUM_PACKET_SIZE= 512,
	INCOMING_BUFFER_SIZE= 1*MEG,
	SCRATCH_BUFFER_SIZE= (16*KILO),

	MAXIMUM_NUMBER_OF_ROOMS= 64,
	FIRST_ROOM_LISTENING_PORT= 6335
};

enum {
	_awaiting_login_user_state,
	_logged_in_state,
	NUMBER_OF_USER_STATES
};

/* ------- parameters structure */
struct user_parameters {
	long host;
	short port;
	short game_search_port;
	char password[16];
};

/* --------- globals */
#ifdef MULTI_ROOM
struct global_room_data *rooms = NULL;
int number_of_rooms = 0;
int current_room = 0;
#else
struct global_room_data room_globals;
#endif
static struct user_parameters parameters;

/* --------- local prototypes */
static void parse_command_arguments(int argc, char *argv[]);
#ifdef MULTI_ROOM
static void bind_rooms(void);
static int init_rooms(void);
static void loop_rooms(void);
#else
static void run_server(void);
#endif
static int handle_new_client(int fd, int set_size);
static void read_data_from_userd_socket(int fd, struct circular_queue userd_incoming);
static void read_data_from_game_search_socket(int fd, struct circular_queue game_search_incoming);
static void read_data_from_client_socket(int fd, fd_set *testfds_write);
static boolean valid_remote_host(unsigned long host);
static boolean handle_userd_data(struct circular_queue *incoming);
static int connect_to_user_daemon(short room_port, short room_identifier);
static int connect_to_game_search_engine(long l_roomd_id);
static boolean handle_userd_packet(struct room_packet_header *header);
static void check_for_dropped_players(fd_set *readfds);
static int idle_room(fd_set *readfds, int set_size);
static void process_pending_app_types(void);
static boolean send_room_status(void);
static boolean update_player_room_list(struct room_packet_header *header);
static boolean handle_game_search_data(struct circular_queue *incoming);
static boolean handle_game_search_packet(struct room_packet_header *header);
static boolean handle_gs_query_response_packet(struct room_packet_header * header);
static boolean handle_rs_player_query_response_packet(struct room_packet_header *header);
static boolean handle_rs_player_info_reply_packet(struct room_packet_header *header);
static boolean handle_rs_player_application_type_packet(struct room_packet_header *header);

/* --------- code */

int main(
	int argc,
	char *argv[])
{
#ifndef MULTI_ROOM
	memset(&room_globals, 0, sizeof(struct global_room_data));
	room_globals.userd_socket= NONE;
	room_globals.game_search_socket = NONE;
	room_globals.room_port= NONE;
	room_globals.room_identifier= NONE;
	room_globals.seconds_before_update= DEFAULT_SECONDS_BEFORE_UPDATE;
	room_globals.maximum_players_per_room= DEFAULT_MAXIMUM_PLAYERS_PER_ROOM;
	room_globals.ticks_at_last_connection_attempt = 0l;
	room_globals.ticks_at_last_room_update = 0l;
	room_globals.game_search_initial_connect = TRUE;
	room_globals.ready_for_search = FALSE;
	room_globals.launch_time = time(NULL);
	room_globals.pending_app_types = NULL;
	room_globals.available_room_count = -1;
#endif	

	initialize_games_log();

	parse_command_arguments(argc, argv);

#ifdef MULTI_ROOM
	init_rooms();
	loop_rooms();
#else
	run_server();
#endif

	return 0;
}

void change_room_motd(
	char *motd)
{
	strncpy(room_globals.motd, motd, ROOM_MAXIMUM_MOTD_SIZE);
	room_globals.motd[ROOM_MAXIMUM_MOTD_SIZE - 1] = '\0';
	
	return;
}

/* --------- local code */
static void parse_command_arguments(
	int argc, 
	char *argv[])
{
	char host_name[256];
	int opt;
	struct hostent *hostdata;
	
	strcpy(host_name, get_userd_host());
	parameters.port= atoi(get_userd_room_port());
	parameters.game_search_port = DEFAULT_GAME_SEARCH_PORT;
	strcpy(parameters.password, "login");

	// parse command line options..
	while((opt= getopt(argc, argv, "dh:r:p:m:")) != -1)
	{
		switch(opt)
		{
			case 'r':
				strcpy(parameters.password, optarg);
				break;
				
#ifndef MULTI_ROOM
			case 'p':
				parameters.port= atoi(optarg);
				break;
#endif
				
			case 'h':
				strcpy(host_name, optarg);
				break;

#ifndef MULTI_ROOM
			case 'm':
				room_globals.maximum_players_per_room= atoi(optarg);
				break;
#endif

			case '?':
				exit(0);
				break;
				
			case ':':
				break;
		}
	}

	hostdata= gethostbyname(host_name);
	if(hostdata)
	{
		if(hostdata->h_addrtype != AF_INET)
		{
			exit(1);
		} else {
			struct in_addr *address;
			
			address= (struct in_addr *) *hostdata->h_addr_list;
			
			parameters.host= ntohl(address->s_addr);
		}
	} else {
		exit(1);
	}

	return;
}

#ifndef MULTI_ROOM
static void run_server(
	void)
{
	int server_socket;

	server_socket= socket(PF_INET, SOCK_STREAM, 0);
	if(server_socket!=NONE) 
	{
		int set_size= 0, i;
		struct sockaddr_in server_address;
		boolean bound= FALSE;

		server_address.sin_family= AF_INET;
		server_address.sin_addr.s_addr= htonl(INADDR_ANY);
		server_address.sin_port= 0;

		signal(SIGPIPE, SIG_IGN);

		for (i= 0; !bound && (i < MAXIMUM_NUMBER_OF_ROOMS); i++)
		{
			server_address.sin_port= htons(FIRST_ROOM_LISTENING_PORT+i);
			if (bind(server_socket, (struct sockaddr *)&server_address, sizeof(server_address))==0)
			{
				bound= TRUE;
			}
		}

		if (bound)
		{
			boolean done= FALSE;
			socklen_t length;

			length= sizeof(struct sockaddr_in);
			getsockname(server_socket, (struct sockaddr *) &server_address, &length);
			assert(length==sizeof(struct sockaddr_in));

			room_globals.room_port= ntohs(server_address.sin_port);
			room_globals.userd_socket= connect_to_user_daemon(room_globals.room_port, NONE);
			if(room_globals.userd_socket != NONE)
			{
				struct circular_queue userd_incoming;
				struct circular_queue game_search_incoming;
			
				if(!allocate_circular_queue("Userd incoming", &userd_incoming, INCOMING_BUFFER_SIZE))
				{
					exit(0);
				}

				if (!allocate_circular_queue("gs incoming", &game_search_incoming, INCOMING_BUFFER_SIZE))
				{
					exit(-1);
				}

				listen(server_socket, MAXIMUM_OUTSTANDING_REQUESTS);

				FD_ZERO(&room_globals.readfds);
				FD_SET(server_socket, &room_globals.readfds);
				FD_SET(room_globals.userd_socket, &room_globals.readfds);
				set_size= MAX(server_socket+1, room_globals.userd_socket+1);

				time(&room_globals.last_userd_ping);

				while(!done) 
				{
					fd_set testfds_read;
					fd_set testfds_write;
					int result;
					struct timeval timeout;					

					testfds_read= room_globals.readfds;
					build_client_writefds(&testfds_write);
					
					timeout.tv_sec= SECONDS_TO_WAIT_ON_SELECT;
					timeout.tv_usec= MICROSECONDS_TO_WAIT_ON_SELECT;
					result= select(set_size, &testfds_read, &testfds_write, (fd_set *) 0, &timeout);

					if((result<0) && (errno != EINTR))
					{
						exit(1);
					} 
					else if(result==0)
					{
						set_size= idle_room(&room_globals.readfds, set_size);
					}
					else
					{
						int fd;
						int current_set_size = set_size;

						for (fd = 0; fd < current_set_size; fd++) 
						{
							if (FD_ISSET(fd, &testfds_read)) 
							{
								if (fd == server_socket)
									set_size = handle_new_client(fd, set_size);
								else if (fd == room_globals.userd_socket)
									read_data_from_userd_socket(fd, userd_incoming);
								else if (fd == room_globals.game_search_socket)
									read_data_from_game_search_socket(fd, game_search_incoming);
								else
									read_data_from_client_socket(fd, &testfds_write);
							}

							if(FD_ISSET(fd, &testfds_write))
							{
								struct client_data *client;

								client= find_client_from_socket(fd);
								if(client) 
								{
									boolean remove_client= FALSE;
		
									remove_client= send_outgoing_data(client);
									if(remove_client) 
									{
										closedown_client(client);
										delete_client(client);
										FD_CLR(fd, &room_globals.readfds);
										close(fd);
									} 
								}
								else
								{
								}
							}
						}

						set_size= idle_room(&room_globals.readfds, set_size);
					}
					
				}
				
				free_circular_queue(&userd_incoming);
			} else {
			}
		}
		close(server_socket);
	}
}
#endif

static int handle_new_client(int fd, int set_size)
{
	struct sockaddr_in remote_address;
	int client_socket;
	socklen_t client_len;
	long host;
	short port;

	errno = 0;
	client_len = sizeof(struct sockaddr_in);
	client_socket = accept(fd, (struct sockaddr *) &remote_address, &client_len);

	if (client_socket != NONE)
	{
		host = ntohl(remote_address.sin_addr.s_addr);
		port = ntohs(remote_address.sin_port);

		if (!room_globals.logged_into_userd || !valid_remote_host(host)) {
			close(client_socket);
		} else if (current_client_count() >= room_globals.maximum_players_per_room + ROOM_PLAYER_LIMIT_SLOP) {
			refuse_client_room_full(client_socket);
			close(client_socket);
		} else if (!add_client(client_socket, host, port)) {
			close(client_socket);
		} else {
			int flags = fcntl(client_socket, F_GETFL, 0);

			fcntl(client_socket, F_SETFL, O_NONBLOCK | flags);
			FD_SET(client_socket, &room_globals.readfds);
			set_size = MAX(set_size, client_socket + 1);
		}
	}

  return set_size;
}


static void read_data_from_userd_socket(int fd, struct circular_queue userd_incoming)
{
	int nread;
	boolean remove_userd = FALSE;

	ioctl(fd, FIONREAD, &nread);
	if (!nread) {
		remove_userd = TRUE;
	} else {
		char buffer[INCOMING_BUFFER_SIZE];

		if (nread >= sizeof(buffer))
			nread = sizeof(buffer) - 1;
		nread = read(fd, buffer, nread);

		if (copy_data_into_circular_queue(buffer, nread, &userd_incoming) == FALSE) {
			remove_userd = TRUE;
		} else {
			remove_userd = handle_userd_data(&userd_incoming);
		}
	}

	if (remove_userd)
	{
		FD_CLR(fd, &room_globals.readfds);
		close(fd);
		room_globals.userd_socket = NONE;
	}
}


static void read_data_from_game_search_socket(int fd, struct circular_queue game_search_incoming)
{
	int nread;
	boolean remove_game_search = FALSE;

	ioctl(fd, FIONREAD, &nread);
	if (!nread) {
		remove_game_search = TRUE;
	} else {
		char buffer[64*KILO];
		
		if (nread >= sizeof(buffer))
			nread = sizeof(buffer) -1;
		nread = read(fd, buffer, nread);
		
		if (copy_data_into_circular_queue(buffer, nread, &game_search_incoming) == FALSE) {
			remove_game_search = TRUE;
		} else {
			remove_game_search = handle_game_search_data(&game_search_incoming);
		}
	}

	if (remove_game_search)
	{
		FD_CLR(fd, &room_globals.readfds);
		close(fd);
		room_globals.game_search_socket = NONE;
	}
}


static void read_data_from_client_socket(int fd, fd_set *testfds_write)
{
	int nread;
	struct client_data *client;

	client = find_client_from_socket(fd);
	if (client) 
	{
		boolean remove_client = FALSE;
		boolean client_has_outgoing_data = FALSE;

		ioctl(fd, FIONREAD, &nread);
		if (!nread) {
			remove_client = TRUE;
		} else {
			char buffer[SCRATCH_BUFFER_SIZE];

			if (nread >= sizeof(buffer)) {
				remove_client = TRUE;
			} else {
				nread = read(fd, buffer, nread);
				
				if (copy_data_into_client_buffer(client, buffer, nread) == FALSE) {
					remove_client = TRUE;
				} else {
					remove_client = handle_client_data(client, &client_has_outgoing_data);
				}
			}
		}

		if (remove_client) {
			closedown_client(client);
			delete_client(client);
			FD_CLR(fd, &room_globals.readfds);
			FD_CLR(fd, testfds_write);
			close(fd);
		} else if (client_has_outgoing_data) {
			FD_SET(fd, testfds_write);
		}
	}
}


#define	TEN_MINUTES		600
#define	SD_BOTH			2

static int idle_room(
	fd_set *readfds,
	int set_size)
{
	time_t current_time;

	time(&current_time);

	check_for_dropped_players(readfds);
	idle_games();

	if (current_time-room_globals.last_userd_ping>TEN_MINUTES)
	{
		if (room_globals.userd_socket!=NONE) 
		{
			shutdown(room_globals.userd_socket, SD_BOTH);
			FD_CLR(room_globals.userd_socket, readfds);
			//pause();
		}

		room_globals.userd_socket= NONE;
		time(&room_globals.last_userd_ping);
	}

	if((room_globals.userd_socket==NONE || 
		room_globals.game_search_socket == NONE) && 
		(machine_tick_count() - room_globals.ticks_at_last_connection_attempt) >
		(60 * MACHINE_TICKS_PER_SECOND))
	{
		if (room_globals.userd_socket == NONE)
		{
			room_globals.userd_socket= connect_to_user_daemon(room_globals.room_port, room_globals.room_identifier);
			if(room_globals.userd_socket==NONE)
			{
			} else {
				FD_SET(room_globals.userd_socket, readfds);
				set_size= MAX(set_size, room_globals.userd_socket+1);
			}
		}

		if (room_globals.game_search_socket == NONE && !room_globals.game_search_initial_connect)
		{
			room_globals.game_search_socket = connect_to_game_search_engine(room_globals.room_identifier);
			if (room_globals.game_search_socket == NONE)
			{
			}
			else
			{
				FD_SET(room_globals.game_search_socket, readfds);
				set_size = MAX(set_size, room_globals.game_search_socket + 1);
			}
		}
		room_globals.ticks_at_last_connection_attempt= machine_tick_count();	
	}
	
	if (room_globals.userd_socket != NONE &&
		(machine_tick_count() - room_globals.ticks_at_last_room_update) >
		(room_globals.seconds_before_update * MACHINE_TICKS_PER_SECOND))
	{
		char buffer[512];
		short length;
	
		length= build_short_room_status_packet(buffer);
		send_room_packet(room_globals.userd_socket, buffer, length);
		room_globals.ticks_at_last_room_update= machine_tick_count();
	}

	if (room_globals.ready_for_search)
	{
		if  (room_globals.game_search_socket == NONE && room_globals.game_search_initial_connect)
		{
			room_globals.game_search_initial_connect = FALSE;
			room_globals.game_search_socket = connect_to_game_search_engine(room_globals.room_identifier);
			if (room_globals.game_search_socket == NONE)
			{
			}
			else
			{
				FD_SET(room_globals.game_search_socket, readfds);
				set_size = MAX(set_size, room_globals.game_search_socket + 1);
			}
		}
	}

	/* process any pending app types we may have */
	process_pending_app_types();

	return set_size;
}

// Myrd - new goodies
static void process_pending_app_types(
	void)
{
	struct client_data *client;
	struct application_type_data *app_type, *next_app_type, *prev_app_type = NULL;
	boolean remove;

	/* scan through any pending app types, to see if we have a match yet */
	for (app_type = room_globals.pending_app_types; app_type; app_type = next_app_type)
	{
		remove = FALSE;
		next_app_type = app_type->next;
		// try to find this guy
		for (client = first_client; client; client = client->next)
		{
			if (client->host == app_type->host && !strcmp(app_type->username, client->name))
			{
				// ok, we should be pretty certain that is our guy.. set his app type
				client->application_type = app_type->application_type;
				// and update his room list
				if (room_globals.available_room_count > 0)
				{
					send_room_list_to_client(client, room_globals.available_rooms, room_globals.available_room_count);
				}
				remove = TRUE;
				break;
			}
		}

		if (app_type->time_stamp + 3*TIME_IN_MINUTES < get_current_time())
		{
			remove = TRUE; // this thing is too old, kill it
		}

		if (remove)
		{
			if (prev_app_type)
			{
				prev_app_type->next = next_app_type;
			}
			else
			{
				room_globals.pending_app_types = next_app_type;
			}
			free(app_type);
		}
		else 
		{
			prev_app_type = app_type;
		}
	}
}
// Myrd - end

static boolean valid_remote_host(
	unsigned long host)
{
	return TRUE;
}

static int connect_to_game_search_engine(long l_room_id)
{
	int sock;
	boolean success= 0;
	
	sock = socket(PF_INET, SOCK_STREAM, 0);
	if (sock != NONE)
	{
		struct sockaddr_in local_address;
		
		local_address.sin_family = AF_INET;
		local_address.sin_addr.s_addr = htonl(INADDR_ANY);
		local_address.sin_port = 0;

		if (bind(sock, (struct sockaddr *)&local_address, sizeof(local_address)) != -1)
		{
			struct sockaddr_in server_address;

			// game search engine and user authentication must be on same machine
			server_address.sin_family = AF_INET;
			server_address.sin_addr.s_addr = htonl(parameters.host);
			server_address.sin_port = htons(parameters.game_search_port);
			if (connect(sock, (struct sockaddr *)&server_address, sizeof(server_address)) == 0)
			{
				short length;
				char buffer[512];

				length = build_gs_login_packet(buffer, l_room_id);
				success = send_game_search_packet(sock, buffer, length);
			}
			else
			{
			}
		}
		else
		{
		}		
	}
	else
	{
	}

	if (!success)
	{
		if (sock != NONE)
		{
			close(sock);
			sock = NONE;
		}
	}

	return sock;
}

static int connect_to_user_daemon(
	short room_port,
	short room_identifier)
{
	int s;
	boolean success= FALSE;
	
	s= socket(PF_INET, SOCK_STREAM, 0);
	if(s!=NONE) 
	{
		struct sockaddr_in local_address;

		local_address.sin_family= AF_INET;
		local_address.sin_addr.s_addr= htonl(INADDR_ANY);
		local_address.sin_port= 0;

		if(bind(s, (struct sockaddr *) &local_address, 
			sizeof(local_address))!=-1) 
		{
			struct sockaddr_in server_address;

			server_address.sin_family= AF_INET;
			server_address.sin_addr.s_addr= htonl(parameters.host); /* we don't care where we go. */
			server_address.sin_port= htons(parameters.port); /* no port required. */

			if(connect(s, (struct sockaddr *) &server_address, sizeof(server_address))==0)
			{
				short length;
				char buffer[512];
			
				length= build_rs_login_packet(buffer, room_port, room_identifier, parameters.password);
				success= send_room_packet(s, buffer, length);
			}
		} else {
		}
	}
	
	if(!success)
	{
		if(s != NONE)
		{
			close(s);
			s= NONE;
		}
	}
	
	return s;
}

static boolean handle_userd_data(
	struct circular_queue *incoming)
{
	boolean disconnect_client= FALSE;
	char packet[64*KILO];
	struct room_packet_header *header;
	boolean byteswap= FALSE;
	boolean disconn_from_error = FALSE;

#ifdef little_endian
	byteswap= TRUE;
#endif

	header= (struct room_packet_header *) packet;
	while(!disconnect_client && get_next_packet_from_queue(incoming, header,
		sizeof(packet), byteswap, &disconn_from_error, /* is_room_packet = */ TRUE))
	{
		if (!byteswap_room_packet((char *) header, FALSE))
			disconnect_client= TRUE;
		else
			disconnect_client= handle_userd_packet(header);
	}

	if (disconn_from_error == TRUE)
	{
		disconnect_client = TRUE;
	}

	return disconnect_client;
}

static boolean handle_game_search_data(struct circular_queue *incoming)
{
	boolean disconnect_client = FALSE;
	char packet[64 * KILO]; 
	struct room_packet_header *header;
	boolean byteswap = FALSE;
	boolean disconn_from_error = FALSE;

#ifdef little_endian
	byteswap= TRUE;
#endif

	header = (struct room_packet_header *) packet;
	while(!disconnect_client && get_next_packet_from_queue(incoming, header,
		sizeof(packet), byteswap, &disconn_from_error, /* is_room_packet = */ TRUE))
	{
		byte_swap_game_search_packet((char *) header, FALSE);
		disconnect_client= handle_game_search_packet(header);
	}

	if (disconn_from_error == TRUE)
	{
		disconnect_client = TRUE;
	}

	return disconnect_client;
}

static boolean handle_userd_packet(
	struct room_packet_header *header)
{
	boolean disconnect= FALSE;
	
	switch(header->type)
	{
		case _rs_ping_packet:
			time(&room_globals.last_userd_ping);
			break;

		case _rs_rank_update_packet:
			{
				char * p;
				struct rs_rank_update_packet * packet;

				p = (char *)header;
				p += sizeof(struct room_packet_header);

				// ALAN Begin: assignment from incompatible pointer type
			//	packet = (struct rs_caste_breakpoint_packet *)p;
				packet = (struct rs_rank_update_packet *)p;
				// ALAN End
				memcpy(&room_globals.caste_breakpoints, &packet->caste_breakpoints, sizeof(struct caste_breakpoint_data));
				memcpy(&room_globals.overall_rank_data, &packet->overall_rank_data, sizeof(struct overall_ranking_data));
			}
			break;

		case _rs_public_announcement_packet:
			{
				char * message;

				message = (char *)header;
				message += sizeof(struct room_packet_header);
				send_room_message_to_everyone(message);
			}
			break;

		case _rs_global_message_packet:
			{
				struct rs_global_message_packet * packet;
				char * message;
				struct client_data * client;

				message = (char *)header;
				message += sizeof(struct room_packet_header);
				packet = (struct rs_global_message_packet *)message;
				message += sizeof(struct rs_global_message_packet);

				client = find_client_from_player_id(packet->player_id);
				send_room_message_to_player(client, message);
			}
			break;

		case _rs_login_successful_packet:
			{
				struct rs_login_successful_packet *packet= (struct rs_login_successful_packet *) ((byte *) header+sizeof(struct room_packet_header));
				char *url_for_update= extract_url_for_update_from_rs_login_successful_packet((char *) header);
				char *motd= extract_motd_from_rs_login_successful_packet((char *) header);

				room_globals.flags= 0;
				room_globals.logged_into_userd= TRUE;
				room_globals.supported_application_flags= packet->supported_application_flags;
				room_globals.room_identifier= packet->identifier;
				room_globals.ranked_room= packet->ranked;
				room_globals.tournament_room = packet->tournament_room;
				strcpy(room_globals.url_for_version_update, url_for_update);
				strcpy(room_globals.motd, motd);
				memcpy(&room_globals.caste_breakpoints, &packet->caste_breakpoints, sizeof(struct caste_breakpoint_data));
				room_globals.ready_for_search = TRUE;
			}
			break;
			
		case _rs_login_failure_packet:
			disconnect= TRUE;
			break;

#ifndef BN2_DEMOVERSION
		case _rs_player_information_packet:
			{
				struct rs_player_information_packet * packet = (struct rs_player_information_packet *) ((byte *) header+sizeof(struct room_packet_header));
				char * p = (char *)header;
				p += sizeof(struct room_packet_header) + sizeof(struct rs_player_information_packet);

				set_client_player_information(packet->player_id, packet->buddies, packet->order, packet->country_code, packet->player_is_admin, packet->player_is_bungie_employee, packet->account_is_kiosk, p);
			}
			break;
#endif
			
		case _rs_client_ranking_packet:
			{
				struct rs_client_ranking_packet *packet= (struct rs_client_ranking_packet *) ((byte *) header+sizeof(struct room_packet_header));

				if (set_client_ranking(packet->player_id, packet->flags, &packet->stats) == FALSE)
				{
					disconnect = TRUE;
				}
			}
			break;

		case _rs_motd_changed_packet:
			change_room_motd(extract_motd_from_rs_motd_changed_packet((char *) header));
			break;
			
		case _rs_send_status_packet:
			disconnect= !send_room_status();
			break;
			
		case _rs_room_list_packet:
			disconnect = update_player_room_list(header);
			break;
			
		case _rs_update_buddy_response_packet:
			{
				struct buddy_entry * buddies;
				struct rs_update_buddy_response_packet * packet;
				long player_id;
				char * p;

				p = (char *)header;
				p += sizeof(struct room_packet_header);

				packet = (struct rs_update_buddy_response_packet *)p;

				player_id = packet->player_id;
				buddies = packet->buddies;

				update_client_buddy_list(player_id, buddies);
				disconnect = FALSE;
			}
			break;

		case _rs_update_order_status_packet:
			{
				struct rs_update_order_status_packet * packet;
				struct order_member order_members[STEFANS_MAXIMUM_ORDER_MEMBERS];
				short index, num_members;
				char * p;

				p = (char *)header;
				p += sizeof(struct room_packet_header);

				packet = (struct rs_update_order_status_packet *)p;
				
				memset(order_members, 0, STEFANS_MAXIMUM_ORDER_MEMBERS * sizeof(struct order_member));
				num_members= MIN(STEFANS_MAXIMUM_ORDER_MEMBERS, packet->member_count);
				
				p += sizeof(struct rs_update_order_status_packet);
				for(index=0; index<num_members; index++)
				{
					order_members[index].player_id = ((struct order_member *)(p))->player_id;
					order_members[index].online = ((struct order_member *)(p))->online;
					p += sizeof(struct order_member);
				}

				send_order_update_to_player(packet->player_id, packet->member_count, &order_members[0]);
				disconnect = FALSE;
			}
			break;

		case _rs_player_query_response_packet:
			disconnect = handle_rs_player_query_response_packet(header);
			break;

		case _rs_player_info_reply_packet:
			disconnect = handle_rs_player_info_reply_packet(header);
			break;

		case _rs_player_application_type_packet:
			disconnect = handle_rs_player_application_type_packet(header);
			break;
			
		default:
			break;
	}
	
	return disconnect;
}

// Myrd - we got an information packet about a potential player's application type
static boolean handle_rs_player_application_type_packet(
	struct room_packet_header *header)
{
	struct application_type_data *app_type;
	struct rs_player_application_type_packet *packet;
	char *p, *username;

	p = (char *)header;
	p += sizeof(struct room_packet_header);

	packet = (struct rs_player_application_type_packet *)p;
	username = p + sizeof(struct rs_player_application_type_packet);

	// check if we already have a pending app type for this guy..
	for (app_type = room_globals.pending_app_types; app_type; app_type = app_type->next)
		if (app_type->host == packet->host && !strcmp(app_type->username, username))
			break;

	if (!app_type) // if not, allocate a new entry
	{
		CREATE(app_type, struct application_type_data, 1);
		app_type->next = room_globals.pending_app_types;
		room_globals.pending_app_types = app_type;
	}

	// ..and set it's fields
	app_type->time_stamp = packet->time_stamp;
	app_type->host = packet->host;
	app_type->application_type = packet->application_type;
	strncpy(app_type->username, username, MAXIMUM_METASERVER_USERNAME);
	app_type->username[MAXIMUM_METASERVER_USERNAME] = '\0';

	// try to process it right away
	process_pending_app_types();

	return FALSE;
}
// Myrd - end

static boolean handle_rs_player_info_reply_packet(
	struct room_packet_header * header)
{
	struct rs_player_info_reply_packet * packet;
	char * p;

	p = (char *)header;
	p += sizeof(struct room_packet_header);

	packet = (struct rs_player_info_reply_packet *)p;

	send_player_info_packet_to_player(packet);
	return FALSE;
}

static boolean handle_rs_player_query_response_packet(
	struct room_packet_header * header)
{
	boolean disconnect = FALSE;
	struct client_data * client;
	struct player_query_response_header * qh;
	struct player_query_response_segment * qs;
	int index;
	int packet_length= 0;
	char buffer[SCRATCH_BUFFER_SIZE];
	char * p;

	p = (char *)header;
	p += sizeof(struct room_packet_header);
	qh = (struct player_query_response_header *)p;

	client = find_client_from_player_id(qh->player_id);
	if (client == NULL)
		return FALSE;

	switch (qh->type)
	{
	case _player_search_query:
		packet_length = start_building_player_search_list_packet(buffer);

		p += sizeof(struct player_query_response_header);
		qs = (struct player_query_response_segment *)p;

		for (index = 0; index < qh->number_of_responses; index++)
		{
			packet_length = add_player_data_to_search_packet(buffer, qs->aux_data.room_id, 
				&qs->aux_data, qs->player_data);
			p += sizeof(struct player_query_response_segment);
			qs = (struct player_query_response_segment *)p;
		}
		break;

	case _order_query:
		packet_length = start_building_order_list_packet(buffer);

		p += sizeof(struct player_query_response_header);
		qs = (struct player_query_response_segment *)p;

		for (index = 0; index < qh->number_of_responses; index++)
		{
			packet_length = add_player_data_to_order_list_packet(buffer, qs->aux_data.room_id, 
				&qs->aux_data, qs->player_data);
			p += sizeof(struct player_query_response_segment);
			qs = (struct player_query_response_segment *)p;
		}
		break;

	case _buddy_query:
		packet_length = start_building_buddy_list_packet(buffer);
		
		p += sizeof(struct player_query_response_header);
		qs = (struct player_query_response_segment *)p;

		for (index = 0; index < qh->number_of_responses; index++)
		{
			packet_length = add_player_data_to_buddy_list_packet(buffer, qs->room_id, 
				&qs->aux_data, qs->player_data);
			p += sizeof(struct player_query_response_segment);
			qs = (struct player_query_response_segment *)p;
		}
		break;
	}

	send_message_to_player_in_room(GET_CLIENT_PLAYER_ID(client), buffer, packet_length);

	return disconnect;	
}

static boolean handle_game_search_packet(struct room_packet_header *header)
{
	boolean disconnect= FALSE;
	
	switch(header->type)
	{
	case _gs_query_response_packet:
		disconnect = handle_gs_query_response_packet(header);
		break;
			
	default:
		break;
	}
	
	return disconnect;
}

static boolean handle_gs_query_response_packet(struct room_packet_header * header)
{
	struct client_data * p_client;
	struct gs_query_response_header * packet_header;
	struct gs_query_response_segment * p_seg;
	char * psz = (char *)header;
	int n_index, n_length;
	char buffer[65536];

	memset(buffer, 0, sizeof(buffer));

	psz += sizeof(struct room_packet_header);
	packet_header = (struct gs_query_response_header *)psz;
	psz += sizeof(struct gs_query_response_header);
	p_seg = (struct gs_query_response_segment *)psz;

	p_client = find_client_from_player_id(packet_header->player_id);
	if (!p_client)
	{
		return FALSE;
	}

	n_length = start_building_game_list_pref_packet(buffer);

	p_seg = (struct gs_query_response_segment *)psz;
	for (n_index = 0; n_index < packet_header->number_of_responses; n_index++)
	{
		byte_swap_game_data((char *)&p_seg->game);
		n_length = add_game_data_to_packet(buffer, &p_seg->aux_data, &p_seg->game, p_seg->game_data_length);
		psz += p_seg->segment_length;
		p_seg = (struct gs_query_response_segment *)psz;
	}

	send_message_to_player_in_room(GET_CLIENT_PLAYER_ID(p_client), buffer, n_length);

	return FALSE;
}

static boolean send_room_status(
	void)
{
	char buffer[2048];
	short length;
	
	length= build_room_status_packet(buffer);
	return send_room_packet(room_globals.userd_socket, buffer, length);
}

static boolean update_player_room_list(
	struct room_packet_header *header)
{
	struct player_room_list_data *rooms= (struct player_room_list_data *) ((char *) header+sizeof(struct room_packet_header));
	short room_count= (header->length-sizeof(struct room_packet_header))/sizeof(struct player_room_list_data);

	if (sizeof(struct room_packet_header)+(room_count*sizeof(struct player_room_list_data))!=header->length)
		return TRUE;

	// Myrd - save the list for future use
	room_globals.available_room_count = room_count;
	memset(room_globals.available_rooms, 0, sizeof(struct player_room_list_data)*MAXIMUM_ROOMS);
	if (room_count > 0)
		memcpy(room_globals.available_rooms, rooms, room_count*sizeof(struct player_room_list_data));
	// Myrd - end

	if (send_room_list_to_all_clients(rooms, room_count) == FALSE)
	{
		return TRUE;
	}
	
	return FALSE;
}

static void check_for_dropped_players(
	fd_set *readfds)
{
	struct client_data *client= find_first_client();

	while(client)
	{
		struct client_data *next= find_next_client(client);
		int client_socket= get_client_socket(client);
	
		if(ping_client(client))
		{
			closedown_client(client);
			delete_client(client);
			FD_CLR(client_socket, readfds);
			close(client_socket);
		}
		client= next;
	}
	
	return;
}

#ifdef MULTI_ROOM
// loop through all rooms, handling anything those rooms need to do
static void loop_rooms(void)
{
	int result;
	struct timeval timeout;					
	fd_set testfds_read;
	fd_set testfds_write;
	boolean done = FALSE;
	
	while (!done)
	{
		for (current_room = 0; current_room < number_of_rooms; current_room++)
		{
			testfds_read = rooms[current_room].readfds;
			build_client_writefds(&testfds_write);

			timeout.tv_sec= SECONDS_TO_WAIT_ON_SELECT;
			timeout.tv_usec= MICROSECONDS_TO_WAIT_ON_SELECT;

			result = select(rooms[current_room].set_size, &testfds_read, &testfds_write, (fd_set *) 0, &timeout);

			if (result < 0 && errno != EINTR)
			{
				puts("Error: select() returned negative value! Exiting...\n");
				exit(1);
			} 
			else if (result != 0)
			{
				int fd;
				int current_set_size = rooms[current_room].set_size;

				for (fd = 0; fd < current_set_size; fd++) 
				{
					if (FD_ISSET(fd, &testfds_read)) 
					{
						if (fd == rooms[current_room].server_socket)
						{
							rooms[current_room].set_size = handle_new_client(fd, rooms[current_room].set_size);
						}
						else if (fd == rooms[current_room].userd_socket)
						{
							read_data_from_userd_socket(fd, rooms[current_room].userd_incoming);
						}
						else if (fd == rooms[current_room].game_search_socket)
						{
							read_data_from_game_search_socket(fd, rooms[current_room].game_search_incoming);
						}
						else
						{
							read_data_from_client_socket(fd, &testfds_write);
						}
					}

					if (FD_ISSET(fd, &testfds_write))
					{
						struct client_data *client = find_client_from_socket(fd);

						if (client) 
						{
							boolean remove_client = FALSE;

							remove_client = send_outgoing_data(client);
							if (remove_client) 
							{
								closedown_client(client);
								delete_client(client);
								FD_CLR(fd, &rooms[current_room].readfds);
								close(fd);
							} 
						}
					}
				}
			}
			rooms[current_room].set_size = idle_room(&rooms[current_room].readfds, rooms[current_room].set_size);
		}
	}
}

// bind each room to its proper port
static void bind_rooms(void)
{
	int port = FIRST_ROOM_LISTENING_PORT;
	int length = sizeof(struct sockaddr_in);

	signal(SIGPIPE, SIG_IGN);

	for (current_room = 0; current_room < number_of_rooms; current_room++)
	{
		rooms[current_room].server_address.sin_family = AF_INET;
		rooms[current_room].server_address.sin_addr.s_addr = htonl(INADDR_ANY);
		rooms[current_room].server_address.sin_port = 0;
		rooms[current_room].server_socket = socket(PF_INET, SOCK_STREAM, 0);
		do
		{
			rooms[current_room].server_address.sin_port = htons(port);
			rooms[current_room].room_port = port;
			port++;
		} while (bind(rooms[current_room].server_socket,
			(struct sockaddr *)&rooms[current_room].server_address,
			sizeof(rooms[current_room].server_address)));

//		printf("New room (ID: %d) listening on port %d.\n",
//			rooms[current_room].room_identifier, rooms[current_room].room_port);
		
		getsockname(rooms[current_room].server_socket, (struct sockaddr *) &rooms[current_room].server_address, &length);
		assert(length==sizeof(struct sockaddr_in));

		listen(rooms[current_room].server_socket, MAXIMUM_OUTSTANDING_REQUESTS);
		FD_ZERO(&rooms[current_room].readfds);
		FD_SET(rooms[current_room].server_socket, &rooms[current_room].readfds);

		rooms[current_room].userd_socket = connect_to_user_daemon(rooms[current_room].room_port, NONE);
		if (rooms[current_room].userd_socket != NONE)
		{
			if (!allocate_circular_queue("Userd incoming", &rooms[current_room].userd_incoming, INCOMING_BUFFER_SIZE))
			{
				exit(0);
			}
			if (!allocate_circular_queue("gs incoming", &rooms[current_room].game_search_incoming, INCOMING_BUFFER_SIZE))
			{
				exit(-1);
			}

			listen(rooms[current_room].server_socket, MAXIMUM_OUTSTANDING_REQUESTS);

			FD_ZERO(&rooms[current_room].readfds);
			FD_SET(rooms[current_room].server_socket, &rooms[current_room].readfds);
			FD_SET(rooms[current_room].userd_socket, &rooms[current_room].readfds);
			rooms[current_room].set_size =
				MAX(rooms[current_room].server_socket+1, rooms[current_room].userd_socket+1);

			time(&rooms[current_room].last_userd_ping);
		}
	}
}

// parse room file, count rooms, allocate memory for each room, then call bind_rooms()
static int init_rooms(void)
{
	FILE *file;
	char name_list[256];
	int room_identifier, ranked_room, country_code;
	int minimum_caste, maximum_caste, tournament_room;

	if (!(file = fopen(get_rooms_list_file(), "r")))
	{
		printf("No rooms list file found! Exiting...\n");
		exit(1);
	}
	
	// first, we count the rooms
	number_of_rooms = 0;
	while (fscanf(file, "%s %d %d %d %d %d %d \n", name_list, &room_identifier, &ranked_room,
			&country_code, &minimum_caste, &maximum_caste, &tournament_room) != EOF)
	{
		number_of_rooms++;
	}

	if (number_of_rooms < 1)
	{
		printf("No valid entries found in rooms list file! Exiting...\n");
		exit(1);
	}

	// allocate memory for our rooms
	CREATE(rooms, struct global_room_data, number_of_rooms);

	// now, scan through the file again and actually fill the data into the structures
	rewind(file);
	for (current_room = 0; current_room < number_of_rooms; current_room++)
	{
		fscanf(file, "%s %d %d %d %d %d %d \n", name_list, &room_identifier, &ranked_room,
			&country_code, &minimum_caste, &maximum_caste, &tournament_room);
		rooms[current_room].room_identifier = room_identifier;
		rooms[current_room].supported_application_flags = get_supported_application_flags_from_name_list(name_list);
		rooms[current_room].ranked_room = ranked_room;
		rooms[current_room].tournament_room = tournament_room;
		rooms[current_room].country_code = country_code;
		rooms[current_room].min_caste = minimum_caste;
		rooms[current_room].max_caste = maximum_caste;
		rooms[current_room].launch_time = time(NULL);
		rooms[current_room].the_first_game = NULL;
		rooms[current_room].global_game_id = 0;
		rooms[current_room].the_first_client = NULL;
		rooms[current_room].ticks_at_last_connection_attempt = 0l;
		rooms[current_room].ticks_at_last_room_update = 0l;
		rooms[current_room].game_search_initial_connect = TRUE;
		rooms[current_room].ready_for_search = FALSE;
		rooms[current_room].maximum_players_per_room= DEFAULT_MAXIMUM_PLAYERS_PER_ROOM;
		rooms[current_room].userd_socket= NONE;
		rooms[current_room].game_search_socket = NONE;
		rooms[current_room].room_port= NONE;
		rooms[current_room].seconds_before_update= DEFAULT_SECONDS_BEFORE_UPDATE;
		rooms[current_room].launch_time = time(NULL);
		rooms[current_room].pending_app_types = NULL;
		rooms[current_room].available_room_count = -1;
	}
	fclose(file);

	// ok, we're done reading from the file, time to bind and such
	bind_rooms();
	
	return number_of_rooms;
}
#endif
